

module("luci.statistics.rrdtool", package.seeall)

require("luci.statistics.datatree")
require("luci.statistics.rrdtool.colors")
require("luci.statistics.i18n")
require("luci.model.uci")
require("luci.util")
require("luci.sys")

local fs = require "nixio.fs"


Graph = luci.util.class()

function Graph.__init__( self, timespan, opts )

	opts = opts or { }

	local uci = luci.model.uci.cursor()
	local sections = uci:get_all( "luci_statistics" )
	opts.timespan  = timespan       or sections.rrdtool.default_timespan or 900
	opts.rrasingle = opts.rrasingle or ( sections.collectd_rrdtool.RRASingle == "1" )
	opts.host      = opts.host      or sections.collectd.Hostname        or luci.sys.hostname()
	opts.width     = opts.width     or sections.rrdtool.image_width      or 400
	opts.rrdpath   = opts.rrdpath   or sections.collectd_rrdtool.DataDir or "/tmp/rrd"
	opts.imgpath   = opts.imgpath   or sections.rrdtool.image_path       or "/tmp/rrdimg"
	opts.rrdpath   = opts.rrdpath:gsub("/$","")
	opts.imgpath   = opts.imgpath:gsub("/$","")
	self.colors = luci.statistics.rrdtool.colors.Instance()
	self.tree   = luci.statistics.datatree.Instance(opts.host)
	self.i18n   = luci.statistics.i18n.Instance( self )
	self.args = {
		"-a", "PNG",
		"-s", "NOW-" .. opts.timespan,
		"-w", opts.width
	}
	self.opts = opts
end

function Graph._mkpath( self, plugin, plugin_instance, dtype, dtype_instance )
	local t = self.opts.host .. "/" .. plugin
	if type(plugin_instance) == "string" and plugin_instance:len() > 0 then
		t = t .. "-" .. plugin_instance
	end
	t = t .. "/" .. dtype
	if type(dtype_instance) == "string" and dtype_instance:len() > 0 then
		t = t .. "-" .. dtype_instance
	end
	return t
end

function Graph.mkrrdpath( self, ... )
	return string.format( "%s/%s.rrd", self.opts.rrdpath, self:_mkpath( ... ) )
end

function Graph.mkpngpath( self, ... )
	return string.format( "%s/%s.%i.png", self.opts.imgpath, self:_mkpath( ... ), self.opts.timespan )
end

function Graph.strippngpath( self, path )
	return path:sub( self.opts.imgpath:len() + 2 )
end

function Graph._forcelol( self, list )
	if type(list[1]) ~= "table" then
		return( { list } )
	end
	return( list )
end

function Graph._rrdtool( self, def, rrd )
	local dir = def[1]:gsub("/[^/]+$","")
	fs.mkdirr( dir )
	local cmdline = "rrdtool graph"
	for i, opt in ipairs(self.args) do
		table.insert( def, 1 + i, opt )
	end
	for i, opt in ipairs(def) do
		opt = opt .. ""    -- force string

		if rrd then
			opt = opt:gsub( "{file}", rrd )
		end

		if opt:match("[^%w]") then
			cmdline = cmdline .. " '" .. opt .. "'"
		else
			cmdline = cmdline .. " " .. opt
		end
	end
	local rrdtool = io.popen( cmdline )
	rrdtool:close()
end

function Graph._generic( self, opts, plugin, plugin_instance, dtype, index )
	local defs = { }
	local _args         = { }
	local _sources	    = { }
	local _stack_neg    = { }
	local _stack_pos    = { }
	local _longest_name = 0
	local _has_totals   = false
	local _ti	    = table.insert
	local _sf	    = string.format
	function _tif( list, fmt, ... )
		table.insert( list, string.format( fmt, ... ) )
	end
	function __def(source)

		local inst = source.sname
		local rrd  = source.rrd
		local ds   = source.ds

		if not ds or ds:len() == 0 then ds = "value" end

		_tif( _args, "DEF:%s_avg=%s:%s:AVERAGE", inst, rrd, ds )

		if not self.opts.rrasingle then
			_tif( _args, "DEF:%s_min=%s:%s:MIN", inst, rrd, ds )
			_tif( _args, "DEF:%s_max=%s:%s:MAX", inst, rrd, ds )
		end

		_tif( _args, "CDEF:%s_nnl=%s_avg,UN,0,%s_avg,IF", inst, inst, inst )
	end
	function __cdef(source)

		local prev
		if source.flip then
			prev = _stack_neg[#_stack_neg]
		else
			prev = _stack_pos[#_stack_pos]
		end
		if not prev or source.overlay then
			_tif( _args, "CDEF:%s_stk=%s_nnl", source.sname, source.sname )
		else
			_tif( _args, "CDEF:%s_stk=%s_nnl,%s_stk,+", source.sname, source.sname, prev )
		end
		if source.flip then
			_tif( _args, "CDEF:%s_neg=%s_stk,-1,*", source.sname, source.sname )
			if not source.overlay then
				_ti( _stack_neg, source.sname )
			end
		elseif not source.overlay then
			_ti( _stack_pos, source.sname )
		end
		if source.total then
			_tif( _args,
				"CDEF:%s_avg_sample=%s_avg,UN,0,%s_avg,IF,sample_len,*",
				source.sname, source.sname, source.sname
			)

			_tif( _args,
				"CDEF:%s_avg_sum=PREV,UN,0,PREV,IF,%s_avg_sample,+",
				source.sname, source.sname, source.sname
			)
		end
	end
	function __cdef_totals()
		if _has_totals then
			_tif( _args, "CDEF:mytime=%s_avg,TIME,TIME,IF", _sources[1].sname    )
			_ti(  _args, "CDEF:sample_len_raw=mytime,PREV(mytime),-"             )
			_ti(  _args, "CDEF:sample_len=sample_len_raw,UN,0,sample_len_raw,IF" )
		end
	end
	function __line(source)

		local line_color
		local area_color
		local legend
		local var
		if type(source.color) == "string" then
			line_color = source.color
			area_color = self.colors:from_string( line_color )
		elseif type(opts.colors[source.name:gsub("[^%w]","_")]) == "string" then
			line_color = opts.colors[source.name:gsub("[^%w]","_")]
			area_color = self.colors:from_string( line_color )
		else
			area_color = self.colors:random()
			line_color = self.colors:to_string( area_color )
		end
		area_color = self.colors:to_string( self.colors:faded( area_color ) )
		if source.flip then
			var = "neg"
		else
			var = "stk"
		end
		legend = _sf( "%-" .. _longest_name .. "s", source.title )
		if not source.noarea then
			_tif( _args, "AREA:%s_%s#%s", source.sname, var, area_color )
		end
		_tif( _args, "LINE%d:%s_%s#%s:%s",
			source.noarea and 2 or 1,
			source.sname, var, line_color, legend )
	end
	function __gprint(source)

		local numfmt = opts.number_format or "%6.1lf"
		local totfmt = opts.totals_format or "%5.1lf%s"
		if not self.opts.rrasingle then
			_tif( _args, "GPRINT:%s_min:MIN:\tMin\\: %s", source.sname, numfmt )
		end
		_tif( _args, "GPRINT:%s_avg:AVERAGE:\tAvg\\: %s", source.sname, numfmt )
		if not self.opts.rrasingle then
			_tif( _args, "GPRINT:%s_max:MAX:\tMax\\: %s", source.sname, numfmt )
		end
		if source.total then
			_tif( _args, "GPRINT:%s_avg_sum:LAST:(ca. %s Total)\\l", source.sname, totfmt )
		else
			_tif( _args, "GPRINT:%s_avg:LAST:\tLast\\: %s\\l", source.sname, numfmt )
		end
	end
	local data_types

	if dtype then
		data_types = { dtype }
	else
		data_types = opts.data.types or { }
	end

	if not ( dtype or opts.data.types ) then
		if opts.data.instances then
			for k, v in pairs(opts.data.instances) do
				_ti( data_types, k )
			end
		elseif opts.data.sources then
			for k, v in pairs(opts.data.sources) do
				_ti( data_types, k )
			end
		end
	end
	for i, dtype in ipairs(data_types) do

		local data_instances

		if not opts.per_instance then
			if type(opts.data.instances) == "table" and type(opts.data.instances[dtype]) == "table" then
				data_instances = opts.data.instances[dtype]
			else
				data_instances = self.tree:data_instances( plugin, plugin_instance, dtype )
			end
		end

		if type(data_instances) ~= "table" or #data_instances == 0 then data_instances = { "" } end
		for i, dinst in ipairs(data_instances) do
			local dname = dtype

			if dinst:len() > 0 then
				dname = dname .. "_" .. dinst
			end
			local data_sources = { "value" }

			if type(opts.data.sources) == "table" then
				if type(opts.data.sources[dname]) == "table" then
					data_sources = opts.data.sources[dname]
				elseif type(opts.data.sources[dtype]) == "table" then
					data_sources = opts.data.sources[dtype]
				end
			end
			for i, dsource in ipairs(data_sources) do

				local dsname  = dtype .. "_" .. dinst:gsub("[^%w]","_") .. "_" .. dsource
				local altname = dtype .. "__" .. dsource
				local dopts = { }

				if type(opts.data.options) == "table" then
					if type(opts.data.options[dsname]) == "table" then
						dopts = opts.data.options[dsname]
					elseif type(opts.data.options[altname]) == "table" then
						dopts = opts.data.options[altname]
					elseif type(opts.data.options[dname]) == "table" then
						dopts = opts.data.options[dname]
					elseif type(opts.data.options[dtype]) == "table" then
						dopts = opts.data.options[dtype]
					end
				end
				_ti( _sources, {
					rrd      = dopts.rrd     or self:mkrrdpath( plugin, plugin_instance, dtype, dinst ),
					color    = dopts.color   or self.colors:to_string( self.colors:random() ),
					flip     = dopts.flip    or false,
					total    = dopts.total   or false,
					overlay  = dopts.overlay or false,
					noarea   = dopts.noarea  or false,
					title    = dopts.title   or nil,
					ds       = dsource,
					type     = dtype,
					instance = dinst,
					index    = #_sources + 1,
					sname    = ( #_sources + 1 ) .. dtype
				} )
				_sources[#_sources].title = self.i18n:ds( _sources[#_sources] )
				if _sources[#_sources].title:len() > _longest_name then
					_longest_name = _sources[#_sources].title:len()
				end
				if _sources[#_sources].total then
					_has_totals = true
				end
			end
		end
	end
	local instances = { "" }

	if opts.per_instance then
		instances = self.tree:data_instances( plugin, plugin_instance, _sources[1].type )
	end
	for i, instance in ipairs(instances) do
		_ti( _args, "-t" )
		_ti( _args, self.i18n:title( plugin, plugin_instance, _sources[1].type, instance, opts.title ) )
		_ti( _args, "-v" )
		_ti( _args, self.i18n:label( plugin, plugin_instance, _sources[1].type, instance, opts.vlabel ) )
		if opts.rrdopts then
			for i, o in ipairs(opts.rrdopts) do _ti( _args, o ) end
		end
		for i, source in ipairs(_sources) do
			if opts.per_instance then
				source.instance = instance
				source.rrd      = self:mkrrdpath( plugin, plugin_instance, source.type, instance )
			end

			__def( source )
		end
		__cdef_totals()
		for i, source in ipairs(_sources) do
			__cdef( _sources[1 + #_sources - i] )
		end
		for i, source in ipairs(_sources) do
			__line( source )
			__gprint( source )
		end
		_ti( _args, 1, self:mkpngpath( plugin, plugin_instance, index .. instance ) )
		_ti( defs, _args )
		_args         = { }
		_stack_pos    = { }
		_stack_neg    = { }
	end

	return defs
end

function Graph.render( self, plugin, plugin_instance, is_index )

	dtype_instances = dtype_instances or { "" }
	local pngs = { }
	local plugin_def = "luci.statistics.rrdtool.definitions." .. plugin
	local stat, def = pcall( require, plugin_def )

	if stat and def and type(def.rrdargs) == "function" then
		local _images = { }
		for i, opts in ipairs( self:_forcelol( def.rrdargs( self, plugin, plugin_instance, nil, is_index ) ) ) do
			if not is_index or not opts.detail then
				_images[i] = { }
				local diagrams = self:_generic( opts, plugin, plugin_instance, nil, i )
				for j, def in ipairs( diagrams ) do
					_images[i][j] = def[1]
					self:_rrdtool( def )
				end
			end
		end
		for y = 1, #_images[1] do
			for x = 1, #_images do
				table.insert( pngs, _images[x][y] )
			end
		end
	end

	return pngs
end
